# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

import multiprocessing as mp
import numpy as np
from numba import jit

from pyrasta.tools import _gdal_temp_dataset, _return_raster, _clone_gdal_dataset
from pyrasta.tools.mapping import GDAL_TO_NUMPY
from pyrasta.utils import split_into_chunks
from tqdm import tqdm

try:
    from osgeo import gdal
except ImportError:
    import gdal

OP_WINDOW_SIZE = 1000


@jit(nopython=True, nogil=True)
def get_xy_block_windows(window_size, raster_x_size, raster_y_size):
    """ Get xy block window coordinates

    Description
    -----------
    Get xy block window coordinates depending
    on raster size and window size
    Used specifically for calculations. All windows are
    retrieved, whichever the size.

    Parameters
    ----------
    window_size: (int, int)
        size of window to read within raster as (width, height)
    raster_x_size: int
        raster's width
    raster_y_size: int
        raster's height

    Yields
    -------
    Window coordinates: tuple
        4-element tuple returning the coordinates of the window within the raster
    """

    for y in range(0, raster_y_size, window_size[1]):
        ysize = min(window_size[1], raster_y_size - y)
        for x in range(0, raster_x_size, window_size[0]):
            xsize = min(window_size[0], raster_x_size - x)

            yield x, y, xsize, ysize


@_return_raster
def _log(raster, out_file):

    out_ds = _clone_gdal_dataset(raster, out_file,
                                 gdal.GetDataTypeByName("float32"))

    for band in range(raster.nb_band):
        for window in get_xy_block_windows((OP_WINDOW_SIZE, OP_WINDOW_SIZE),
                                           raster.x_size,
                                           raster.y_size):

            array = raster._gdal_dataset.GetRasterBand(band + 1).\
                ReadAsArray(*window).astype("float32")
            out_ds.GetRasterBand(band + 1).WriteArray(np.log(array), window[0], window[1])

    # Close dataset
    out_ds = None


@_return_raster
def _log10(raster, out_file):

    out_ds = _clone_gdal_dataset(raster, out_file,
                                 gdal.GetDataTypeByName("float32"))

    for band in range(raster.nb_band):
        for window in get_xy_block_windows((OP_WINDOW_SIZE, OP_WINDOW_SIZE),
                                           raster.x_size,
                                           raster.y_size):

            array = raster._gdal_dataset.GetRasterBand(band + 1).\
                ReadAsArray(*window).astype("float32")
            out_ds.GetRasterBand(band + 1).WriteArray(np.log10(array), window[0], window[1])

    # Close dataset
    out_ds = None


@_return_raster
def _op(raster1, out_file, raster2, op_type):
    """ Basic arithmetic operations

    """
    out_ds = _clone_gdal_dataset(raster1, out_file,
                                 data_type=gdal.GetDataTypeByName('float32'))

    for band in range(1, raster1.nb_band + 1):

        for window in get_xy_block_windows((OP_WINDOW_SIZE, OP_WINDOW_SIZE),
                                           raster1.x_size,
                                           raster1.y_size):
            arrays = []
            for src in [raster1, raster2]:
                try:
                    arrays.append(src._gdal_dataset.GetRasterBand(
                        band).ReadAsArray(*window).astype("float32"))
                except AttributeError:
                    arrays.append(src)

            if op_type == "add":
                result = arrays[0] + arrays[1]
            elif op_type == "sub":
                result = arrays[0] - arrays[1]
            elif op_type == "rsub":
                result = arrays[1] - arrays[0]
            elif op_type == "mul":
                result = arrays[0] * arrays[1]
            elif op_type == "pow":
                result = arrays[0] ** arrays[1]
            elif op_type == "rpow":
                result = arrays[1] ** arrays[0]
            elif op_type == "truediv":
                result = np.full(arrays[0].shape, raster1.no_data)
                if not np.isscalar(arrays[1]):
                    result[arrays[1] != 0] = \
                        arrays[0][arrays[1] != 0] / arrays[1][arrays[1] != 0]
                else:
                    if arrays[1] != 0:
                        result = arrays[0] / arrays[1]
            elif op_type == "rtruediv":
                result = np.full(arrays[0].shape, raster1.no_data)
                result[arrays[0] != 0] = arrays[1] / arrays[0][arrays[0] != 0]
            else:
                result = None

            if np.isscalar(arrays[1]):
                result[arrays[0] == raster1.no_data] = raster1.no_data
            else:
                result[(arrays[0] == raster1.no_data) | (arrays[1] == raster2.no_data)] = \
                    raster1.no_data

            out_ds.GetRasterBand(band).WriteArray(result, window[0], window[1])

    # Close dataset
    out_ds = None


@_return_raster
def _raster_calculation(raster_class, out_file, gdal_driver, sources,
                        fhandle, window_size, input_type, output_type,
                        no_data, nb_processes, chunksize, description):
    """ Calculate raster expression

    """
    if not hasattr(window_size, "__getitem__"):
        window_size = (window_size, window_size)

    master_raster = sources[0]
    window_gen = ([src._gdal_dataset.ReadAsArray(*w).astype(GDAL_TO_NUMPY[input_type]) for src in
                   sources] for w in get_xy_block_windows(window_size,
                                                          master_raster.x_size,
                                                          master_raster.y_size))
    width = int(master_raster.x_size /
                window_size[0]) + min(1, master_raster.x_size % window_size[0])
    height = int(master_raster.y_size /
                 window_size[1]) + min(1, master_raster.y_size % window_size[1])

    # Initialization
    is_first_run = True
    y = 0

    if description:
        iterator = tqdm(split_into_chunks(window_gen, width),
                        total=height,
                        desc=description)
    else:
        iterator = split_into_chunks(window_gen, width)

    for win_gen in iterator:

        with mp.Pool(processes=nb_processes) as pool:
            list_of_arrays = list(pool.map(fhandle,
                                           win_gen,
                                           chunksize=chunksize))

        result = np.concatenate(list_of_arrays, axis=list_of_arrays[0].ndim - 1)

        if is_first_run:
            if result.ndim == 2:
                nb_band = 1
            else:
                nb_band = result.shape[0]

            out_ds = _gdal_temp_dataset(out_file,
                                        gdal_driver,
                                        master_raster._gdal_dataset.GetProjection(),
                                        master_raster.x_size,
                                        master_raster.y_size, nb_band,
                                        master_raster.geo_transform,
                                        output_type,
                                        no_data)

            is_first_run = False

        if nb_band == 1:
            out_ds.GetRasterBand(1).WriteArray(result, 0, y)
        else:
            for band in range(nb_band):
                out_ds.GetRasterBand(band + 1).WriteArray(result[band, :, :],
                                                          0, y)

        y += window_size[1]

    # Close dataset
    out_ds = None



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""

from functools import wraps

from pyrasta import GDAL_DEFAULT_DRIVER
from pyrasta.io_.files import RasterTempFile

try:
    from osgeo import gdal
except ImportError:
    import gdal


def driver_authorizes_creation(gdal_driver):

    return True if 'DCAP_CREATE' in gdal_driver.GetMetadata().keys() else False


def _return_raster(function):
    @wraps(function)
    def return_raster(raster, *args, **kwargs):
        try:
            if driver_authorizes_creation(raster._gdal_driver):
                gdal_driver = raster._gdal_driver
            else:
                gdal_driver = GDAL_DEFAULT_DRIVER
            with RasterTempFile(gdal_driver.GetMetadata()['DMD_EXTENSION']) as out_file:
                function(raster, out_file.path, *args, **kwargs)
                new_raster = raster.__class__(out_file.path)
        except AttributeError:
            gdal_driver = [arg for arg in args if isinstance(arg, gdal.Driver)][0]
            args = [arg for arg in args if not isinstance(arg, gdal.Driver)]
            with RasterTempFile(gdal_driver.GetMetadata()['DMD_EXTENSION']) as out_file:
                try:
                    function(raster, out_file.path, gdal_driver, *args, **kwargs)
                except TypeError:
                    function(raster, out_file.path, *args, **kwargs)
                new_raster = raster(out_file.path)
        finally:
            new_raster._temp_file = out_file

        return new_raster
    return return_raster


def _gdal_temp_dataset(out_file, gdal_driver, projection, x_size, y_size,
                       nb_band, geo_transform, data_type, no_data):
    """ Create gdal temporary dataset

    """
    try:
        out_ds = gdal_driver.Create(out_file, x_size, y_size, nb_band, data_type)
    except RuntimeError:
        out_ds = GDAL_DEFAULT_DRIVER.Create(out_file, x_size,
                                            y_size, nb_band, data_type)

    out_ds.SetGeoTransform(geo_transform)
    out_ds.SetProjection(projection)
    _set_no_data(out_ds, no_data)

    return out_ds


def _clone_gdal_dataset(raster, out_file, data_type=None):

    if data_type is None:
        data_type = raster.data_type

    return _gdal_temp_dataset(out_file,
                              raster._gdal_driver,
                              raster._gdal_dataset.GetProjection(),
                              raster.x_size,
                              raster.y_size,
                              raster.nb_band,
                              raster.geo_transform,
                              data_type,
                              raster.no_data)


def _set_no_data(gdal_ds, no_data):
    """ Set no data value into gdal dataset

    Description
    -----------

    Parameters
    ----------
    gdal_ds: gdal.Dataset
        gdal dataset
    no_data: list or tuple
        list of no data values corresponding to each raster band

    """
    for band in range(gdal_ds.RasterCount):
        try:
            gdal_ds.GetRasterBand(band + 1).SetNoDataValue(no_data)
        except TypeError:
            pass



# -*- coding: utf-8 -*-

""" Module summary description.

More detailed description.
"""
from pyrasta.crs import srs_from
from pyrasta.utils import gdal_progress_bar

try:
    from osgeo import gdal, ogr
except ImportError:
    import gdal
    import ogr


def _polygonize(raster, filename, band, layer_name,
                field_name, ogr_driver, is_8_connected,
                progress_bar):
    """ Polygonize raster

    Parameters
    ----------
    raster
    filename
    band
    layer_name
    field_name
    ogr_driver
    is_8_connected
    progress_bar

    Returns
    -------

    """
    connectivity = "8CONNECTED=%d" % (8 if is_8_connected else 4)
    dst_ds = ogr_driver.CreateDataSource(filename)
    dst_layer = dst_ds.CreateLayer(layer_name,
                                   geom_type=ogr.wkbPolygon,
                                   srs=srs_from(raster.crs))

    fd = ogr.FieldDefn(field_name, ogr.OFTInteger)
    dst_layer.CreateField(fd)

    callback, callback_data = gdal_progress_bar(progress_bar,
                                                description="Polygonize raster")

    srcband = raster._gdal_dataset.GetRasterBand(band)
    maskband = srcband.GetMaskBand()

    gdal.Polygonize(srcband,
                    maskband,
                    dst_layer,
                    0,
                    [connectivity],
                    callback=callback,
                    callback_data=callback_data)

    dst_ds = None

    return 0
